local locations = require(mod_name .. ".lualib.locations")
local math2d = require('math2d')

local biter_strategy = {}
local biter_strategy_data = {
  colonies = {},
  tracked_structures = {},
}

local shuffle_table = function (tbl)
  for i = #tbl, 2, -1 do
    local j = math.random(i)
    tbl[i], tbl[j] = tbl[j], tbl[i]
  end
  return tbl
end

local random_point_in_box = function(box)
  return {
    x= math.random(box.left_top.x,box.right_bottom.x),
    y= math.random(box.left_top.y,box.right_bottom.y)
  }
end

local explosion = function(surface,position)
  local types = {'blood-explosion-small'}
  local random_type = types[math.random(1,#types)]
  surface.create_entity({
    name=random_type,
    position=position,
    speed=10,
    source=position,
    target=position,
    })
end

local decorate = function(surface,position)
  local types = {'lichen'}
  local random_type = types[math.random(1,#types)]
  surface.create_decoratives({
    check_collision = false,
    decoratives = {
    {
      name=random_type,
      position=position,
      amount=1
    }
    }})
end

local convert_colony_data =  function (data)
  assert(locations.get_area(data.area_name),"biter_strategy: attempt to use invalid area name")
  local new_colony = {
    name = data.name,
    area_name = data.area_name,
    area = locations.get_area(data.area_name),
    reinforcement_sources = data.reinforcement_sources,
    reinforcement_group = nil,
    force = data.force,
    lichen_active = data.lichen_active or false,
    surface = locations.get_main_surface(),
    desired_structures = data.desired_structures,
    allowed_reinforcement_types =data.allowed_reinforcement_types,
    allow_reinforcements = data.allow_reinforcements,
    allow_pollution_absorption = data.allow_pollution_absorption,
    reinforcement_wave_size = data.reinforcement_wave_size,
    next_update = game.ticks_played + math.random(1,100),
    complete = false,
    active = true,
    debug_ids = {},
  }

  if campaign_debug_mode then
    table.insert(new_colony.debug_ids,rendering.draw_rectangle({
      left_top = new_colony.area.left_top,
      right_bottom = new_colony.area.right_bottom,
      color = { r = 1, b = 0.5, g = 0, a = 1 },
      forces = { 'player' },
      surface = new_colony.surface,
    }))
    table.insert(new_colony.debug_ids,rendering.draw_text({
      text = new_colony.name,
      surface = new_colony.surface,
      color = { r = 1, b = 0.5, g = 0, a = 1 },
      forces = { 'player' },
      target = new_colony.area.left_top,
      scale = 2,
    }))
  end

  for _, structure in pairs(data.desired_structures) do
    if biter_strategy_data.tracked_structures[structure.name] == nil then
      biter_strategy_data.tracked_structures[structure.name] = true
    end
  end

  return new_colony
end

local build_next_colony_structure = function(colony)
  for _, desired_structure in pairs(colony.desired_structures) do
    if colony.surface.count_entities_filtered({
      name = desired_structure.name,
      area = colony.area,
    }) < desired_structure.count then
      local build_pos = colony.surface.find_non_colliding_position(
        desired_structure.name,
        random_point_in_box(colony.area),
        5,
        1,
        true
      )
      if build_pos then
        colony.surface.create_entity({
          name = desired_structure.name,
          position = build_pos,
          force = colony.force,
        })
      else
        print("no place for new "..desired_structure.name.." at "..colony.name)
      end
      return
    end
  end
end

local do_messy_sacrifice = function(biter,lichen)
  if lichen then
    decorate(biter.surface,biter.position)
  end
  explosion(biter.surface,biter.position)
  biter.die()
end

local send_group_to_colony = function(group,destination_colony)
  local random_destination = random_point_in_box(destination_colony.area)
  group.set_command({
    type = defines.command.compound,
    structure_type = defines.compound_command.return_last,
    commands = {
      {
        type = defines.command.go_to_location,
        destination = random_destination,
        distraction = defines.distraction.by_damage,
        radius = 0.1,
      },
      {
        type = defines.command.wander,
        distraction = defines.distraction.by_anything,
        --distraction = defines.distraction.by_damage,
        radius = 5,
        wander_in_group = true,
      }
    }
  })
  destination_colony.reinforcement_group = group
end

local create_reinforcement_group_at_colony = function(source_colony,destination_colony)
  print("created free units at "..source_colony.name.." to send to "..destination_colony.name)
  local new_group = destination_colony.surface.create_unit_group({
    position = math2d.bounding_box.get_centre(source_colony.area),
    force = destination_colony.force,
  })
  local random_biter_type
  local art = destination_colony.allowed_reinforcement_types
  for _=1, destination_colony.reinforcement_wave_size do
    random_biter_type = art[math.random(1,#art)]
    --local random_free_point = source_colony.surface.find_non_colliding_position_in_box(
    --  random_biter_type,
    --  source_colony.area,
    --  0.5
    --)
    local random_free_point = source_colony.surface.find_non_colliding_position(
      random_biter_type,
      random_point_in_box(source_colony.area),
      5,
      0.1
    )
    print(random_biter_type.." being spawned at "..serpent.line(random_free_point)..
      " "..source_colony.name.." going to "..destination_colony.name)

    if random_free_point then
      new_group.add_member(source_colony.surface.create_entity({
        name = random_biter_type,
        position = random_free_point,
        force = destination_colony.force,
      }))
    end
  end

  if #new_group.members > 0 then
    send_group_to_colony(new_group,destination_colony)
  end
end

local send_reinforcement_group_from_spawners = function(spawner_list,destination_colony)
  print("stole units from spawner to send to "..destination_colony.name)
  assert(#spawner_list>0,"biter_strategy: trying to take biters from invalid spawners")
  local random_spawner = spawner_list[math.random(1,#spawner_list)]
  if #random_spawner.units > 0 then
    local new_group = destination_colony.surface.create_unit_group({
      force = destination_colony.force,
      position = random_spawner.position,
    })
    for _, unit in pairs(random_spawner.units) do
      unit.release_from_spawner()
      new_group.add_member(unit)
    end
    send_group_to_colony(new_group,destination_colony)
  else
    print("biter_strategy: reinforcement source has spawners but no biters")
  end
end

local update_colony = function(colony)
  local next_update_wait_time
  if colony.surface.count_entities_filtered(
    {
      name = {"biter-spawner","spitter-spawner"},
      area = colony.area,
    }) > 0 then
    print("updated "..colony.name.." it has spawners")
    local spawners = colony.surface.find_entities_filtered({
      name = {"biter-spawner","spitter-spawner"},
      limit=1,
      area = colony.area,
    })
    if #spawners[1].units > 5 then
      print("updated "..colony.name.." it has units at spawner 1")
      --kill all the units
      for _, unit in pairs(spawners[1].units) do
        do_messy_sacrifice(unit,colony.lichen_active)
      end
      build_next_colony_structure(colony)
      next_update_wait_time = 1
    else
      next_update_wait_time = 2
    end
  else --no spawners colony
    print("updated "..colony.name.." has no spawners")
    local r_group = colony.reinforcement_group
    if r_group and r_group.valid then
      print("updated "..colony.name.." it has a reinforcement group")
      -- use them for building if they are in the colony area or wandering
      if math2d.bounding_box.contains_point(colony.area,r_group.position) or
        r_group.state == defines.group_state.wander_in_group then
        print("updated "..colony.name.." its reinforcement group has arrived")
        for _, unit in pairs(r_group.members) do
          do_messy_sacrifice(unit,colony.lichen_active)
        end
        build_next_colony_structure(colony)
        next_update_wait_time = 2
      else
        print("updated "..colony.name.." waiting for reinforcements")
        --wait for them to arrive
        next_update_wait_time = 10
      end
    else
      -- check reinforcement_sources for biters to create a reinforcement group
      print("updated "..colony.name.." has no reinforcement group")
      if colony.allow_reinforcements == false then
        next_update_wait_time = 20
      elseif #colony.reinforcement_sources == 0 then
        print("updated "..colony.name.." is making its own reinforcements")
        --im a spawn point so I make my own biters
        create_reinforcement_group_at_colony(colony,colony)
        next_update_wait_time = 1
      else
        if #colony.reinforcement_sources > 0 then
          shuffle_table(colony.reinforcement_sources)
        end
        for _, source_colony_name in pairs(colony.reinforcement_sources) do
          local source_colony = biter_strategy_data.colonies[source_colony_name]
          assert(source_colony,"biter_strategy: attempt to read invalid source "..source_colony_name.." for colony "..colony.name)
          local source_colony_spawners = source_colony.surface.find_entities_filtered({
            name = {'biter-spawner','spitter-spawner'},
            area = source_colony.area,
          })
          if #source_colony.reinforcement_sources == 0 then
            print("updated "..colony.name.." source is a map edge base")
            --this is a map edge spawn point, so just create biters
            create_reinforcement_group_at_colony(source_colony,colony)
            next_update_wait_time = 5
            break
          elseif #source_colony_spawners > 0 then
            print("updated "..colony.name.." source has spawners ")
            --take some biters from a spawner here
            send_reinforcement_group_from_spawners(source_colony_spawners,colony)
            next_update_wait_time = 5
            break
          else
            print("updated "..colony.name.." source cant help ")
            --there are no reinforcements to get here
            next_update_wait_time = 10
          end
        end
      end
    end
  end
  assert(next_update_wait_time,"biter_strategy: somehow an update did not set next_update_wait_time")
  colony.next_update = game.ticks_played + (next_update_wait_time * 3000)
end

local evaluate_colony = function(colony)
  local complete = true
  for _, desired_structure in pairs(colony.desired_structures) do
    if colony.surface.count_entities_filtered({
      name = desired_structure.name,
      area = colony.area,
    }) < desired_structure.count then
      complete = false
    end
  end
  colony.complete = complete
  print(colony.name.." colony evaluated to be complete")
end

local evaluate_colony_at = function(position)
  for _, colony in pairs(biter_strategy_data.colonies) do
    if math2d.bounding_box.contains_point(colony.area,position) then
      evaluate_colony(colony)
      return
    end
  end
end

biter_strategy.load_data = function(data)
  for _, point in pairs(data) do
    biter_strategy_data.colonies[point.name] = convert_colony_data(point)
  end
end

biter_strategy.update = function()
  local updated_something = false
  for _, colony in pairs(biter_strategy_data.colonies) do
    if colony.active and #colony.desired_structures > 0 and
      colony.complete == false and game.ticks_played > colony.next_update then
      update_colony(colony)
      evaluate_colony(colony)
      updated_something = true
    end
  end
  if updated_something == true then
    print("biter_strategy: updated on tick "..game.ticks_played)
  end
end

biter_strategy.clear_colony = function(colony)
  if colony.reinforcement_group and colony.reinforcement_group.valid then
    colony.reinforcement_group.set_autonomous()
    colony.reinforcement_group = nil
  end
  for _, id in pairs(colony.debug_ids) do
    rendering.destroy(id)
  end
end

biter_strategy.clear_colony_by_name = function(colony_name)
  assert(biter_strategy_data.colonies[colony_name],"biter_strategy: trying to clear an invalid colony "..colony_name)
  biter_strategy.clear_colony(biter_strategy_data.colonies[colony_name])
  biter_strategy_data.colonies[colony_name] = nil
end

biter_strategy.clear_all_colonies = function()
  for _, colony in pairs(biter_strategy_data.colonies) do
    biter_strategy.clear_colony(colony)
  end
  biter_strategy_data.colonies = {}
  biter_strategy_data.tracked_structures = {}
end

biter_strategy.show_debug_to = function(player)
  for _, colony in pairs(biter_strategy_data.colonies) do
    for _, id in pairs(colony.debug_ids) do
      rendering.set_players(id,{player})
      game.print(colony.name)
    end
  end
end

biter_strategy.on_entity_died = function(event)
  if biter_strategy_data.tracked_structures[event.entity.name] then
    print("biter_strategy: tracked entity died")
    evaluate_colony_at(event.entity.position)
  end
end

biter_strategy.init = function()
  global.biter_strategy_data = biter_strategy_data
end

biter_strategy.on_load = function()
  biter_strategy_data = global.biter_strategy_data or biter_strategy_data
end

biter_strategy.events = {
  [defines.events.on_tick] = biter_strategy.update,
  [defines.events.on_entity_died] = biter_strategy.on_entity_died,
}

return biter_strategy